//  (C) Copyright David Abrahams 2000. Permission to copy, use, modify, sell and
//  distribute this software is granted provided this copyright notice appears
//  in all copies. This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//
//  The author gratefully acknowleges the support of Dragon Systems, Inc., in
//  producing this work.

#include "functions.h"
#include "newtypes.h"
#include "singleton.h"
#include "objects.h"

namespace py {

struct Function::TypeObject :
        Singleton<Function::TypeObject, Callable<py::TypeObject<Function> > >
{
    TypeObject() : SingletonBase(&PyType_Type) {}
};

Function::Function()
    : PythonObject(TypeObject::singleton())
{
}

Ptr BoundFunction::create(Ptr target, Ptr fn)
{
    BoundFunction* result = free_list;
    if (result != 0)
    {
        free_list = result->m_free_list_link;
        result->m_target = target;
        result->m_unbound_function = fn;
    }
    else
    {
        result = new BoundFunction(target, fn);
    }
    return Ptr(result, Ptr::new_ref);
}

struct BoundFunction::TypeObject :
    Singleton<BoundFunction::TypeObject, Callable<py::TypeObject<BoundFunction> > >
{
    TypeObject() : SingletonBase(&PyType_Type) {}
    
private: // TypeObject<BoundFunction> hook override
    void dealloc(BoundFunction*) const;
};

BoundFunction::BoundFunction(Ptr target, Ptr fn)
    : PythonObject(TypeObject::singleton()),
      m_target(target),
      m_unbound_function(fn),
      m_free_list_link(0)
{
}

PyObject*
BoundFunction::call(PyObject* args, PyObject* keywords) const
{
    // Build a new tuple which prepends the target to the arguments
    Tuple tail_arguments(Ptr(args, Ptr::borrowed));
    Ptr all_arguments(PyTuple_New(tail_arguments.size() + 1));

    PyTuple_SET_ITEM(all_arguments.get(), 0, m_target.get());
    Py_INCREF(m_target.get());
    for (std::size_t i = 0; i < tail_arguments.size(); ++i)
    {
        PyTuple_SET_ITEM(all_arguments.get(), i + 1, tail_arguments[i].get());
        Py_INCREF(tail_arguments[i].get());
    }

    return PyEval_CallObjectWithKeywords(m_unbound_function.get(), all_arguments.get(), keywords);
}

void BoundFunction::TypeObject::dealloc(BoundFunction* instance) const
{
    instance->m_free_list_link = free_list;
    free_list = instance;
    instance->m_target.reset();
    instance->m_unbound_function.reset();
}

BoundFunction* BoundFunction::free_list;

}
