//  (C) Copyright David Abrahams 2000. Permission to copy, use, modify, sell and
//  distribute this software is granted provided this copyright notice appears
//  in all copies. This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//
//  The author gratefully acknowleges the support of Dragon Systems, Inc., in
//  producing this work.

#ifndef FUNCTIONS_DWA051400_H_
# define FUNCTIONS_DWA051400_H_

# include "pyconfig.h"
# include "wrap_python.h"
# include "pyptr.h"
# include "signatures.h"
# include "caller.h"
# include "base_object.h"

namespace py {

class Function : public PythonObject
{
public:
    Function();
    // Function objects are reasonably rare, so we guess we can afford a virtual table.
    // This cuts down on the number of distinct type objects which need to be defined.
    virtual ~Function() {}
    virtual PyObject* call(PyObject* args, PyObject* keywords) const = 0;

private:
    struct TypeObject;
};

template <class R, class F>
struct NewMemberFunction : Function
{
	typedef F Pmf; // pointer-to-member-function
	
	NewMemberFunction(Pmf pmf)
        : m_pmf(pmf) {}
    
	PyObject* call(PyObject* args, PyObject* keywords) const
        { return Caller<R>::call(m_pmf, args, keywords); }
private:
	const Pmf m_pmf;
};

// A helper function for new_member_function(), below.  Implements the core
// functionality once the return type has already been deduced. R is expected to
// be Type<X>, where X is the actual return type of pmf.
template <class F, class R>
Ptr new_member_function_aux(R, F pmf)
{
    // We can't just use "typename R::Type" below because MSVC (incorrectly) pukes.
    typedef typename R::Type ReturnType;
    return Ptr(new NewMemberFunction<ReturnType, F>(pmf));
}

// Create and return a new member function object wrapping the given
// pointer-to-member function
template <class F>
inline Ptr new_member_function(F pmf)
{
    // Deduce the return type and pass it off to the helper function above
	return new_member_function_aux(return_value(pmf), pmf);
}

class BoundFunction : public PythonObject
{
public:
    static Ptr create(Ptr target, Ptr fn);
    
    BoundFunction(Ptr target, Ptr fn);
    PyObject* call(PyObject*args, PyObject* keywords) const;
private:
    struct TypeObject;
    friend struct TypeObject;
    
    Ptr m_target;
    Ptr m_unbound_function;
    BoundFunction* m_free_list_link;

	static BoundFunction* free_list;
};

}

#endif // FUNCTIONS_DWA051400_H_
