//  (C) Copyright David Abrahams 2000. Permission to copy, use, modify, sell and
//  distribute this software is granted provided this copyright notice appears
//  in all copies. This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//
//  The author gratefully acknowleges the support of Dragon Systems, Inc., in
//  producing this work.

#ifndef INIT_FUNCTION_DWA052000_H_
# define INIT_FUNCTION_DWA052000_H_

# include "pyconfig.h"
# include "functions.h"
# include "signatures.h"

namespace py {

class ExtensionInstance;
class InstanceHolderBase;

class Init;
template <class T> struct Init0;
template <class T, class A1> struct Init1;
template <class T, class A1, class A2> struct Init2;
template <class T, class A1, class A2, class A3> struct Init3;
template <class T, class A1, class A2, class A3, class A4> struct Init4;

template <class T>
struct InitFunction
{
    static Ptr create(Signature0)
        { return Ptr(new Init0<T>); }

    template <class A1>
    static Ptr create(Signature1<A1>)
        { return Ptr(new Init1<T, A1>); }
    
    template <class A1, class A2>
    static Ptr create(Signature2<A1, A2>)
        { return Ptr(new Init2<T, A1, A2>); }

    template <class A1, class A2, class A3>
    static Ptr create(Signature3<A1, A2, A3>)
        { return Ptr(new Init3<T, A1, A2, A3>); }

    template <class A1, class A2, class A3, class A4>
    static Ptr create(Signature4<A1, A2, A3, A4>)
        { return Ptr(new Init4<T, A1, A2, A3, A4>); }
};

class Init : public Function
{
private: // override Function hook
    PyObject* call(PyObject* args, PyObject* keywords) const;
private:
    virtual InstanceHolderBase* create_holder(ExtensionInstance* self, PyObject* tail_args, PyObject* keywords) const = 0;
};

template <class T>
struct Init0 : Init
{
    virtual InstanceHolderBase* create_holder(ExtensionInstance* self, PyObject* args, PyObject* /*keywords*/) const
    {
        if (!PyArg_ParseTuple(args, const_cast<char*>("")))
            throw ErrorAlreadySet();
        return new T(self);
    }
};

template <class T, class A1>
struct Init1 : Init
{
    virtual InstanceHolderBase* create_holder(ExtensionInstance* self, PyObject* args, PyObject* /*keywords*/) const
    {
        PyObject* a1;
        if (!PyArg_ParseTuple(args, const_cast<char*>("O"), &a1))
            throw ErrorAlreadySet();
        return new T(self,
            from_python(a1, Type<A1>())
            );
    }
};

template <class T, class A1, class A2>
struct Init2 : Init
{
    virtual InstanceHolderBase* create_holder(ExtensionInstance* self, PyObject* args, PyObject* /*keywords*/) const
    {
        PyObject* a1;
        PyObject* a2;
        if (!PyArg_ParseTuple(args, const_cast<char*>("OO"), &a1, &a2))
            throw ErrorAlreadySet();
        return new T(self,
            from_python(a1, Type<A1>()),
            from_python(a2, Type<A2>())
            );
    }
};

template <class T, class A1, class A2, class A3>
struct Init3 : Init
{
    virtual InstanceHolderBase* create_holder(ExtensionInstance* self, PyObject* args, PyObject* /*keywords*/) const
    {
        PyObject* a1;
        PyObject* a2;
        PyObject* a3;
        if (!PyArg_ParseTuple(args, const_cast<char*>("OOO"), &a1, &a2, &a3))
            throw ErrorAlreadySet();
        return new T(self,
            from_python(a1, Type<A1>()),
            from_python(a2, Type<A2>()),
            from_python(a3, Type<A3>())
            );
    }
};

template <class T, class A1, class A2, class A3, class A4>
struct Init4 : Init
{
    virtual InstanceHolderBase* create_holder(ExtensionInstance* self, PyObject* args, PyObject* /*keywords*/) const
    {
        PyObject* a1;
        PyObject* a2;
        PyObject* a3;
        PyObject* a4;
        if (!PyArg_ParseTuple(args, "OOOO", &a1, &a2, &a3, &a4))
            throw ErrorAlreadySet();
        return new T(self,
            from_python(a1, Type<A1>()),
            from_python(a2, Type<A2>()),
            from_python(a3, Type<A3>()),
            from_python(a4, Type<A4>())
            );
    }
};

}

#endif // INIT_FUNCTION_DWA052000_H_
